//-----------------------------------------------------------------------------
// Ohjelmointiputka.net "lyks ohjelma" kilpailu
// Tekoly "Plkky"
//
// Sami Luostarinen aka BadHabit
// potaatti@hotmail.com
//
// http://koti.mbnet.fi/badhabit
//-----------------------------------------------------------------------------

#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <dirent.h>

#include "patternsolver.h"

using namespace std;

// CONSTANTS

// Own pattern files for contest
const string output         = "polkky.kir";
const string input          = "polkky.luk";

// File types for pattern files
// which are solved when using -a argument
const string PatternDir     = ".\\patterns\\";  // file dir
const string InPattern      = ".luk";           // input files
const string OutPattern     = ".kir";           // output files

//-----------------------------------------------------------------------------
void WriteAnswer( const string& aFile, vector<string>& aAnswerArray )
//-----------------------------------------------------------------------------
{
    // Writes given string array into file
    fstream outfile( aFile.c_str(), ios::out);
    int count = aAnswerArray.size();
    for( int i = 0; i < count; i++ )
        {
        outfile << aAnswerArray[i] << "\n";
        }
    outfile.close();
}

//-----------------------------------------------------------------------------
void PrintArray( vector<string>& aAnswers )
//-----------------------------------------------------------------------------
{
    // print out an array of strings
    int count = aAnswers.size();
    for( int i = 0; i < count; i++ )
        {
        cout << aAnswers[i] << ",";
        }
}

//-----------------------------------------------------------------------------
bool ReadTask( const string& aFile, vector<string>& aPattern,
               vector<string>& aVerify, int& aRequiredAnswers )
//-----------------------------------------------------------------------------
{
    // read job from file
    fstream infile( aFile.c_str(), ios::in);
    if( !infile.is_open() )
        {
        cout << "Could not open input file!" << endl;
        return false;
        }
    
    // read header
    int count;
    int answers;
    infile >> count >> answers;
    aRequiredAnswers = answers;
    
    // read pattern
    string readin;
    for( int i = 0; i < count; i++ )
        {
        infile >> readin;
        aPattern.push_back( readin );
        }

    // try reading answer also.. for development use
    for( int i = 0; i < answers; i++ )
        {
        readin = "";
        infile >> readin;
        if( readin != "" )
            aVerify.push_back( readin );
        }

    // read succesfull
    infile.close();
    return true;
}

//-----------------------------------------------------------------------------
void FindPatternFiles( vector<string>& aInFiles, vector<string>& aOutFiles )
//-----------------------------------------------------------------------------
{
    // check out for pattern files in directory: [PatternDir]
    cout << "Finding pattern files... ";
    
    DIR *pdir;
    struct dirent *pent;

    if( !( pdir = opendir( PatternDir.c_str() ) ) )
        return;
        
    int count = 0;
    while ( ( pent = readdir( pdir ) ) )
        {
        string file( pent->d_name );
        string::size_type loc = file.find( InPattern, 0 );
        if( loc != string::npos )
            {
            string fullPath = PatternDir + file;
            aInFiles.push_back( fullPath );
            file.replace( loc, InPattern.length(), OutPattern );
            fullPath = PatternDir + file;
            aOutFiles.push_back( fullPath );
            count++;
            }
        }
    cout << "Found " << count << " files." << endl;
    closedir(pdir);
}

//-----------------------------------------------------------------------------
bool VerifyAnswer( vector<string>& aAns, vector<string>& aVerify )
//-----------------------------------------------------------------------------
{
    // Verify that answer is correct
    if( aAns.size() != aVerify.size() )
        return false;

    for( int i = 0; i < aAns.size(); i++ )
        {
        if( aAns[i] != aVerify[i] )
            return false;
        }
    return true;
}

//-----------------------------------------------------------------------------
int main(int argc, char *argv[])
//-----------------------------------------------------------------------------
{
    bool pauseAfterSolve = false;
    
    vector<string> infiles;
    vector<string> outfiles;

    // Check arguments
    if( argc > 1 )
        {
        string arg( argv[1] );
        if( 0 == arg.compare( "-a" ) )
            {
            FindPatternFiles( infiles, outfiles );
            pauseAfterSolve = true;
            }
        }

    // At least solve our "own" file
    if( infiles.size() == 0 )
        {
        cout << "Solving default pattern: " << input << endl;
        infiles.push_back( input );
        outfiles.push_back( output );
        }

    if( infiles.size() != outfiles.size() ) 
        return -1; // fatal error
    
    PatternSolver* solver = new PatternSolver();

    // Solve all wanted patterns
    int solved = 0;
    int count = infiles.size();
    for( int i = 0; i < count; i++ )
        {
        // Read problem
        int required;
        vector<string> pattern;
        vector<string> verify;
        if( !ReadTask( infiles[i], pattern, verify, required ) )
            continue;

        // Solve problem
        solver->SetPattern( pattern, required );
        clock_t ticks = clock();
        if( solver->IterativeSolve() )
            {
            // Write answer
            solved++;
            vector<string> answer;
            solver->GetAnswer( answer );
            WriteAnswer( outfiles[i], answer );

            if( verify.size() > 0 && !VerifyAnswer( answer, verify ) )
                {
                // print wrongly solved cases
                cout << "ERROR (" << infiles[i] << " - " << required << "): ";
                PrintArray( pattern );
                cout << " = ";

                // Print debug log from solving...
                cout << solver->SolveString() << " = ";

                PrintArray( answer );
                cout << "[";
                PrintArray( verify );
                cout << "]";
                cout << endl;
                }
            else if( infiles.size() < 11 )
                {
                // could not verify answer --> print
                cout << "Solved: ";
                PrintArray( pattern );
                cout << " => ";
                PrintArray( answer );
                cout << endl;
                }
            }
        else
            {
            cout << "!! Unable to solve (" << required << "): ";
            PrintArray( pattern );
            cout << endl;
            }

        // Time print
        // cout << "[time:" << (float)( clock() - ticks ) / ( CLK_TCK )<< "]" << endl;
        }


    cout << "Total: " << solved << " / " << count << " solved (" << solved*100/count << "\%)" << endl;
    delete solver;
    if( pauseAfterSolve )
        {
        system("PAUSE");
        }
    return 0;
}
