#include <iostream>
#include "splitsolver.h" 
#include "patternsolver.h"
#include "solverutils.h"

//-----------------------------------------------------------------------------
// Split pattern into two smaller patterns and then try to solve them
// e.g. A1,B2,C3 => A,B,C : 1,2,3
//
bool SplitSolver::Solve()
//-----------------------------------------------------------------------------
{
    iAnswer.clear();
    if( iMaxIteration <= 0 )
        return false;
        
    // Setup solver
    PatternSolver* solver = new PatternSolver();
    solver->SetMaxIteration( iMaxIteration - 1 );

    vector<string> left;
    vector<string> right;

    // Try to solve different split cases
    for( int i = EAlphaStart; i < ESplitEnd; i++ )
        {
        if( !CreateSplitArrays( left, right, (ESplitPattern)i ) )
            continue;
            
        // solve left part
        solver->SetPattern( left, iRequired );
        if( !solver->Solve() )
            continue;
        solver->GetAnswer( left );
        string solvestr = solver->SolveString();
        
        // solve right part
        solver->SetPattern( right, iRequired );
        if( !solver->Solve() )
            continue;
        solver->GetAnswer( right );

        // Solved!
        string splitType;
        NameOfType( i, splitType );
        iSolverStr = "(split:" + splitType + ")[ " + solvestr + " ][ " + solver->SolveString() + " ]";
        delete solver;
        SolverUtils::VectorSum( left, right, iAnswer );
        return true;
        }

    // not solved
    delete solver;
    return false;
}


//-----------------------------------------------------------------------------
bool SplitSolver::CreateSplitArrays( vector<string>& aLeft,
                                     vector<string>& aRight,
                                     ESplitPattern aPattern )
//-----------------------------------------------------------------------------
{
    aLeft.clear();
    aRight.clear();
    
    vector<int> split;
    CalculateSplitLocations( split, aPattern );

    int count = iPattern.size();
    for( int i = 0; i < count; i++ )
        {
        int pos = split[i];
        int len = iPattern[i].length();
        if( pos == 0 || pos >= len )
            return false; // bad split

        aLeft.push_back( iPattern[i].substr( 0, pos ) );
        aRight.push_back( iPattern[i].substr( pos, len - pos ) );
        }
    return true; // ok
}

//-----------------------------------------------------------------------------
void SplitSolver::CalculateSplitLocations( vector<int>& aSplitLen,
                                           ESplitPattern aPattern )
//-----------------------------------------------------------------------------
{
aSplitLen.clear();
int count = iPattern.size();

switch( aPattern )
    {
    // 1 char from begin
    case EAlphaStart:
        {
        for( int i = 0; i < count; i++ )
            {
            int len = min( 1, (int)iPattern[i].length() );
            aSplitLen.push_back( len );
            }
        break;
        }
    // 1 char from the end
    case EAlphaEnd:
        {
        for( int i = 0; i < count; i++ )
            {
            int len = iPattern[i].length() - 1;
            aSplitLen.push_back( max( len, 0 ) );
            }
        break;
        }
    // All number chars from the beginning
    case ENumStart:
        {
        for( int i = 0; i < count; i++ )
            {
            int j = 0;
            while( ( j < iPattern[i].length() ) &&
                   ( iPattern[i][j] >= '0' ) &&
                   ( iPattern[i][j] <= '9' ) )
                {
                j++;
                }
            aSplitLen.push_back( j );
            }
        break;
        }
    // All number chars from the beginning
    case ENumEnd:
        {
        for( int i = 0; i < count; i++ )
            {
            int j = iPattern[i].length() - 1;
            while( ( j >= 0 ) &&
                   ( iPattern[i][j] >= '0' ) &&
                   ( iPattern[i][j] <= '9' ) )
                {
                j--;
                }
            aSplitLen.push_back( max( 0, j ) );
            }
        break;
        }
    // from middle
    case EMiddle:
        {
        for( int i = 0; i < count; i++ )
            {
            int pos = iPattern[i].length() / 2;
            aSplitLen.push_back( pos );
            }
        break;
        }
    // Cumulative split
    case EDownSlide:
        {
        for( int i = 0; i < count; i++ )
            {
            int pos = min( (int)iPattern[i].length(), i + 1);
            aSplitLen.push_back( pos );
            }
        break;
        }
    // Cumulative split (reverse)
    case EUpSlide:
        {
        int first = iPattern[0].length();
        for( int i = 0; i < count; i++ )
            {
            int pos = min( (int)iPattern[i].length(), first - i - 1 );
            pos = max( pos, 0 );
            aSplitLen.push_back( pos );
            }
        break;
        }
    }
}

//-----------------------------------------------------------------------------
void SplitSolver::NameOfType( int aType, string& aTypeStr )
//-----------------------------------------------------------------------------
{
    // type text for debugging use
    switch( aType )
        {
        // 1 char from begin
        case EAlphaStart:
            {
            aTypeStr = "a.start";
            break;
            }
        // 1 char from the end
        case EAlphaEnd:
            {
            aTypeStr = "a.end";
            break;
            }
        // All number chars from the beginning
        case ENumStart:
            {
            aTypeStr = "n.start";
            break;
            }
        // All number chars from the beginning
        case ENumEnd:
            {
            aTypeStr = "n.end";
            break;
            }
        // from middle
        case EMiddle:
            {
            aTypeStr = "mid";
            break;
            }
        // Cumulative split
        case EDownSlide:
            {
            aTypeStr = "cum";
            break;
            }
        // Cumulative split (reverse)
        case EUpSlide:
            {
            aTypeStr = "rcum";
            break;
            }
        }
}
